import pytest
from ejudge import functions
from ejudge.tests.test_io_grader_python import (
    iospec,
    test_run_valid_source,
    test_run_valid_source_with_timeout,
    test_run_valid_source_with_sandbox,
    test_run_valid_source_with_sandbox_and_timeout,
    test_grade_correct,
    test_grade_wrong,
)


@pytest.fixture
def src_ok():
    return (
        'nome = leia_texto("name: ")\n'
        'mostre("hello %s!" % nome)')


@pytest.fixture
def src_bad():
    return (
        'nome = leia_texto("name: ")\n'
        'mostre("hello %s." % nome)')


@pytest.fixture
def lang():
    return 'pytuga'


def test_pytuga_keeps_a_clean_environment(src_ok):
    functions.run(src_ok, ['tuga'], lang='pytuga', sandbox=False)
    assert str(True) == 'True'



from .meta import __version__, __author__
from .registry_class import registry
from .exceptions import BuildError, MissingInputError, EarlyTerminationError
from .functions import run, grade
from . import langs as _langs




from ejudge.build_manager import IntegratedBuildManager, \
    InterpretedLanguageBuildManager
from ejudge.execution_manager import IntegratedExecutionManager, \
    InterpretedLanguageExecutionManager
from ejudge.util import format_traceback


# Python 3.x support
class PythonBuildManager(IntegratedBuildManager):
    """
    Integrated build manager for Python 3.x.
    """

    def syntax_check(self):
        python3_syntax_check(self.source)


class PythonExecutionManager(IntegratedExecutionManager):
    """
    Execution manager for Python source.
    """


# Python 3.x support in an isolated interpreter. This is implemented only for
# debug purposes
class PythonScriptBuildManager(InterpretedLanguageBuildManager):
    """
    Python 3.x builder that executes code in a separate interpreter.
    """

    source_extension = '.py'

    def syntax_check(self):
        python3_syntax_check(self.source)


class PythonScriptExecutionManager(InterpretedLanguageExecutionManager):
    """
    Python 3.x execution in a separate interpreter.
    """

    interpreter_command = 'python3'


# Python 2.7 support
class Python2BuildManager(InterpretedLanguageBuildManager):
    """
    Builds Python 2.7 code.
    """

    source_extension = '.py'

    def syntax_check(self):
        python2_syntax_check(self.source)


class Python2ExecutionManager(InterpretedLanguageExecutionManager):
    """
    Executes Python 2.7 code.
    """

    interpreter_command = 'python2'


# Utilty functions
def python3_syntax_check(source):
    """
    Checks if string of source code is valid Python 3 syntax.
    """

    try:
        compile(source, 'main.py', 'exec')
    except SyntaxError as ex:
        msg = format_traceback(ex, source)
        raise SyntaxError(msg)


def python2_syntax_check(source):
    """
    Checks if string of source code is valid Python 2.7 syntax.
    """



from pprint import pprint
import pytest
from ejudge import functions
from iospec import parse_string, types, SimpleTestCase


@pytest.fixture
def iospec():
    return parse_string(
        'name: <foo>\n'
        'hello foo!\n'
        '\n'
        'name: <bar>\n'
        'hello bar!')


@pytest.fixture
def lang():
    return 'python'


@pytest.fixture
def src_ok():
    return (
        'name = input("name: ")\n'
        'print("hello %s!" % name)'
    )


@pytest.fixture
def src_bad():
    return (
        'name = input("name: ")\n'
        'print(name)'
    )


#
# Test simple io.run() interactions
#
def test_run_valid_source(src_ok, lang, timeout=None, sandbox=False):
    tree = functions.run(src_ok, ['foo'], lang=lang, sandbox=sandbox, timeout=timeout,
                         raises=True)
    try:
        assert len(tree) == 1
        case = tree[0]
        assert isinstance(case, SimpleTestCase)
        assert case[0] == 'name: '
        assert case[1] == 'foo'
        assert case[2] == 'hello foo!'
    except Exception:
        tree.pprint()
        raise


def test_run_valid_source_with_timeout(src_ok, lang):
    test_run_valid_source(src_ok, lang, timeout=1.0)


@pytest.mark.slow
def test_run_valid_source_with_sandbox(src_ok, lang):
    test_run_valid_source(src_ok, lang, sandbox=True)


@pytest.mark.slow
def test_run_valid_source_with_sandbox_and_timeout(src_ok, lang):
    test_run_valid_source(src_ok, lang, sandbox=True, timeout=1.0)


#
# Test io.grade() and check if the feedback is correct
#
def test_grade_correct(iospec, src_ok, lang):
    feedback = functions.grade(src_ok, iospec, lang=lang, sandbox=False)
    assert isinstance(feedback.answer_key, types.TestCase)
    assert isinstance(feedback.testcase, types.TestCase)
    assert feedback.grade == 1
    assert feedback.message is None
    assert feedback.status == 'ok'


def test_grade_wrong(iospec, src_bad, lang):
    feedback = functions.grade(src_bad, iospec, lang=lang, sandbox=False)
    assert feedback.grade == 0
    assert feedback.status == 'wrong-answer'
    assert feedback.title == 'Wrong Answer'


def test_run_from_input_sequence(src_ok, lang):
    tree = functions.run(src_ok, [['foo'], ['bar']], lang=lang, sandbox=False)
    assert len(tree) == 2
    assert tree[0][2] == 'hello foo!'
    assert tree[1][2] == 'hello bar!'


def test_run_from_iospec_input(src_ok, lang):
    case1 = types.SimpleTestCase([types.In('foo'), types.Out('foo')])
    case2 = types.InputTestCase([types.In('bar')])
    inpt = types.IoSpec ([case1, case2])
    tree = functions.run(src_ok, inpt, lang=lang, sandbox=False)
    assert len(tree) == 2
    assert tree[0][2] == 'hello foo!'
    assert tree[1][2] == 'hello bar!'


def test_run_code_with_runtime_error():
    tree = functions.run('1/0', ['foo'], lang='python', sandbox=False)
    lines = tree[0].error_message.splitlines()
    assert lines[0] == 'Traceback (most recent call last)'
    assert lines[1].startswith('  File ')
    assert lines[-1].startswith('ZeroDivisionError')


def test_run_code_with_syntax_error():
    tree = functions.run('bad syntax', ['foo'], lang='python', sandbox=False)
    assert len(tree) == 1
    assert isinstance(tree[0], types.ErrorTestCase)
    # assert tree[0].error_message == (
    #     'Traceback (most recent call last)\n'
    #     '  File "main.py", line 1\n'
    #     '    bad syntax\n'
    #     '             ^\n'
    #     'SyntaxError: invalid syntax\n')
    # assert 'bad syntax' in tree[0].error_message


def test_grade_with_runtime_error():
    iospec = parse_string('foo<bar>\nfoobar')
    feedback = functions.grade('1/0', iospec, lang='python', sandbox=False)
    assert feedback.grade == 0
    assert feedback.title == 'Runtime Error'
    assert feedback.testcase.error_message == (
        'Traceback (most recent call last)\n'
        '  File "main.py", line 1, in <module>\n\n'
        'ZeroDivisionError: division by zero')


def test_run_recursive_function():
    src = '''
def f(x):
    return 1 if x <= 1 else x * f(x - 1)
print(f(5))
'''
    result = functions.run(src, [()], lang='python', sandbox=False)
    assert list(result[0]) == ['120']


def test_run_recursive_function_in_sandbox():
    src = '''
def f(x):
    return 1 if x <= 1 else x * f(x - 1)
print(f(5))
'''
    result = functions.run(src, [()], lang='python', sandbox=True)
    assert list(result[0]) == ['120']




import os
import stat
import tempfile

import subprocess

from ejudge.exceptions import BuildError


class BuildManager:
    """
    Stores information about a program build.
    """

    @classmethod
    def from_json(cls, json):
        return cls(**json)

    def __init__(self, source, is_sandboxed=False, modules=None):
        self.source = source
        self.modules = modules
        self.is_sandboxed = is_sandboxed
        self.is_built = False
        self.is_closed = False

    def build(self):
        """
        Prepares build for the given context.
        """

        self.is_built = True

    def close(self):
        """
        Explicitly deallocate all resources allocated by the BuildManager.
        """

        self.is_closed = True

    def to_json(self):
        """
        A JSON compatible representation of object.
        """

        return self.__dict__.copy()

    def syntax_check(self):
        """
        Raise a SyntaxError if source code syntax is invalid.
        """

    def get_modules(self):
        """
        Return a list of additional modules that should be pre-loaded by
        box.run() in sandboxed mode.
        """

        return (list(self.modules or ())) + [self.__class__.__module__]


class IntegratedBuildManager(BuildManager):
    """
    Build context for Python or other languages that can be integrated into
    the main interpreter.

    These languages can provide a locals and a globals dictionary for passing
    symbols to be used during execution.
    """

    # noinspection PyShadowingBuiltins
    def __init__(self, source, locals=None, globals=None, **kwargs):
        super().__init__(source, **kwargs)
        self.locals = locals
        self.globals = globals


class ExternalProgramBuildManager(BuildManager):
    """
    Build context for compiled programs or programs that run in external
    interpreters. This build context prepares a temporary folder to store all
    necessary files.
    """

    build_path = None
    source_extension = None

    def build(self):
        self.build_path = self.build_tempdir()
        self.prepare_files()
        super().build()

    def build_tempdir(self):
        """
        Creates a temporary directory for storing files.
        """

        # Make tmp directory readable and writable by everyone, if in sandboxed
        # mode.
        temp_dir = tempfile.mkdtemp()
        if self.is_sandboxed:
            os.chmod(
                temp_dir,
                stat.S_IEXEC | stat.S_IREAD | stat.S_IWRITE |
                stat.S_IXOTH | stat.S_IROTH | stat.S_IWOTH |
                stat.S_IXGRP | stat.S_IRGRP | stat.S_IWGRP
            )
        return temp_dir

    def write(self, path, data):
        """
        Write contents of the "data" string into the a "path" relative to the
        working temporary directory.
        """

        if self.build_path is None:
            raise RuntimeError('must create temporary directory first!')

        full_path = os.path.join(self.build_path, path)

        # Save source and make it readable by everyone
        with open(full_path, 'w') as F:
            F.write(data)
        if self.is_sandboxed:
            os.chmod(full_path, stat.S_IREAD | stat.S_IROTH | stat.S_IRGRP)

    def prepare_files(self):
        """
        Saves all necessary files to the temporary directory.
        """

        self.write(self.get_source_filename(), self.source)

    def get_source_extension(self):
        """
        Return the source file default extension.
        """

        if self.source_extension is None:
            raise NotImplementedError(
                'Subclasses must either define the attribute source_extension '
                'or override the .get_source_extension() method.'
            )
        return self.source_extension

    def get_source_filename(self):
        """
        Return the default source file name.
        """

        ext = self.get_source_extension()
        if not ext:
            return 'main'
        else:
            return 'main.' + ext.lstrip('.')


class CompiledLanguageBuildManager(ExternalProgramBuildManager):
    """
    Basic support for languages that require a separate compiling step.
    """

    build_args = None
    executable_name = 'main.exe'

    def prepare_files(self):
        super().prepare_files()

        # Compile and return
        current_path = os.getcwd()
        error_msg = 'compilation is taking too long'
        try:
            os.chdir(self.build_path)
            build_args = self.get_build_args()
            error_msg = subprocess.check_output(build_args, timeout=10)

            # Make executable readable and executable by everyone in sandbox
            # mode
            if self.is_sandboxed:
                os.chmod(
                    self.executable_name,
                    stat.S_IREAD | stat.S_IROTH | stat.S_IRGRP |
                    stat.S_IEXEC | stat.S_IXOTH | stat.S_IXGRP
                )

        except (subprocess.CalledProcessError, TimeoutError):
            raise BuildError(error_msg)
        finally:
            # Cannot return to path in sandboxed mode since the nobody user
            # probably does not have read permissions on the old working
            # directory
            if not self.is_sandboxed:
                os.chdir(current_path)

    def get_build_args(self):
        """
        Return a list with the build args to be passed to the compiler process.
        """

        if self.build_args is None:
            raise NotImplementedError(
                'Subclasses must either define the .build_args attribute or '
                'override the .get_build_args() method.'
            )

        if isinstance(self.build_args, str):
            import shlex
            return shlex.split(self.build_args)

        return list(self.build_args)


class InterpretedLanguageBuildManager(ExternalProgramBuildManager):
    """
    Basic support for scripting languages and interpreter-like execution.
    """


import pytest
from ejudge import functions
from iospec import parse_string, types


#
# Prevent regressions by registering specific interaction bugs
#
def test_less_inputs_than_expected(lang='python', sandbox=False):
    src = 'print(42)'
    iosrc = 'foo<42>\nbar<0>\n42'
    fb = functions.grade(src, iosrc, lang=lang, sandbox=sandbox)
    fb.testcase.pprint()
    fb.answer_key.pprint()
    assert fb.grade == 0
    assert fb.status == 'wrong-answer'


@pytest.mark.slow
def test_less_inputs_than_expected_sandbox():
    test_less_inputs_than_expected(sandbox=True)


def test_less_inputs_than_expected_pytuga():
    test_less_inputs_than_expected(lang='pytuga')



import pytest
from iospec import parse_string, types
from ejudge import functions
from ejudge.tests.test_io_grader_python import (
    iospec,
    test_run_valid_source,
    test_run_valid_source_with_timeout,
    test_run_valid_source_with_sandbox,
    test_run_valid_source_with_sandbox_and_timeout,
    test_grade_correct,
    test_grade_wrong,
    src_ok,
    src_bad
)


@pytest.fixture
def lang():
    return 'python-script'



import pytest
from ejudge import functions
from iospec import parse_string, types
from ejudge.tests.test_io_grader_python import (
    iospec,
    test_run_valid_source,
    test_run_valid_source_with_timeout,
    test_run_valid_source_with_sandbox,
    test_run_valid_source_with_sandbox_and_timeout,
)


@pytest.fixture
def lang():
    return 'gcc'


@pytest.fixture
def src_ok():
    return r"""
#include<stdio.h>

void main() {
    char buffer[100];
    printf("name: ");
    scanf("%s", buffer);
    printf("hello %s!\n", buffer);
}
"""


@pytest.fixture
def src_bad():
    return r"""
#include<stdio.h>

void main() {
    char buffer[100];
    puts("name: ");
    scanf("%s", buffer);
    printf("%s\n", buffer);
}
"""


def test_c_program_with_two_inputs():
    src = r"""
#include<stdio.h>

void main() {
    char name[100], job[200];
    printf("name: ");
    scanf("%s", name);
    printf("job: ");
    scanf("%s", job);
    printf("%s, %s\n", name, job);
}
"""
    result = functions.run(src, ['foo', 'bar'], lang='c', sandbox=False)
    case = result[0]
    print(case)
    result.pprint()
    assert list(case) == ['name: ', 'foo', 'job: ', 'bar', 'foo, bar']



import pytuga
from ejudge.build_manager import IntegratedBuildManager
from ejudge.execution_manager import IntegratedExecutionManager


class PytugaBuildManager(IntegratedBuildManager):
    """
    Build manager for the Pytuga language.
    """

    def syntax_check(self):
        return NotImplemented

    def get_modules(self):
        return super().get_modules() + ['pytuga.lib.forbidden']


class PytugaExecutionManager(IntegratedExecutionManager):
    """
    Execution manager for the Pytugues language.
    """

    def exec(self, globals, locals):
        pytuga.exec(self.source, globals, locals, forbidden=True)



from ejudge.build_manager import CompiledLanguageBuildManager
from ejudge.execution_manager import CompiledLanguageExecutionManager


#
# C compilers
#
# We have manager classes for the two biggest compilers gcc and clang and also
# for tcc, the tiny C compiler, which is a small and fast compiler. In the
# future we might create an IntegratedManager based on clang or tcc or even on
# a C interpreter like ROOT's Cint.
#
class CLanguageBuildManager(CompiledLanguageBuildManager):
    """
    Common functionality for all C-based builders.
    """

    source_extension = '.c'

    def syntax_check(self):
        return NotImplemented


class CLanguageExecutionManager(CompiledLanguageExecutionManager):
    """
    Execution manager for all C-based languages.
    """


class GccBuildManager(CLanguageBuildManager):
    """
    Build executable with GCC.
    """

    build_args = ['gcc', '-lm', 'main.c', '-std=c99', '-o', 'main.exe']


class TccBuildManager(CLanguageBuildManager):
    """
    Build executable with TCC.
    """

    build_args = ['tcc', '-w', '-o', 'main.exe', 'main.c']


class ClangBuildManager(CLanguageBuildManager):
    """
    Build executable with Clang compiler.
    """

    build_args = ['clang', '-lm', 'main.c', '-std=c99', '-o', 'main.exe']


#
# C++ compilers
#
class CppBuildManager(CompiledLanguageBuildManager):
    """
    Base class for C++ builders.
    """

    source_extension = '.cpp'

    def syntax_check(self):
        pass


class GccCppBuildManager(CppBuildManager):
    """
    Build executable with g++
    """

    build_args = ['g++', '-lm', '-o', 'main.cpp', '-o', 'main.exe']


class ClangCppBuildManager(CppBuildManager):
    """
    Build executable from C++ source with clang.
    """

    build_args = ['clang', '-lm', '-o', 'main.cpp', '-o', 'main.exe']



from ejudge.registry_class import registry


# C family languages
registry.register(
    'c',
    'ejudge.langs.c_family.GccBuildManager',
    'ejudge.langs.c_family.CLanguageExecutionManager',
    extensions=['.c'],
    aliases=['gcc', 'C', 'cc']
)
registry.register(
    'tcc',
    'ejudge.langs.c_family.TccBuildManager',
    'ejudge.langs.c_family.CLanguageExecutionManager',
)
registry.register(
    'clang',
    'ejudge.langs.c_family.ClangBuildManager',
    'ejudge.langs.c_family.CLanguageExecutionManager',
)
registry.register(
    'c++',
    'ejudge.langs.c_family.GccCppBuildManager',
    'ejudge.langs.c_family.CLanguageExecutionManager',
    extensions=['.cpp', '.c++'],
    aliases=['cpp', 'g++', 'C++']
)
registry.register(
    'clang-c++',
    'ejudge.langs.c_family.ClangCppBuildManager',
    'ejudge.langs.c_family.CLanguageExecutionManager',
    aliases=['clang++']
)


# Python family
registry.register(
    'python',
    'ejudge.langs.python_family.PythonBuildManager',
    'ejudge.langs.python_family.PythonExecutionManager',
    extensions=['.py', 'py3'],
    aliases=['python3', 'python3x', 'python3.x', 'py', 'py3', 'py3x', 'py3.x']
)
registry.register(
    'python2',
    'ejudge.langs.python_family.Python2BuildManager',
    'ejudge.langs.python_family.Python2ExecutionManager',
    extensions=['.py2'],
    aliases=['python27', 'python2.7', 'py2', 'py27']
)
registry.register(
    'python-script',
    'ejudge.langs.python_family.PythonScriptBuildManager',
    'ejudge.langs.python_family.PythonScriptExecutionManager',
)


# Other scripting languages
registry.register(
    'ruby',
    'ejudge.langs.ruby.RubyBuildManager',
    'ejudge.langs.ruby.RubyExecutionManager',
    extensions=['.rb'],
    aliases=['rb'],
)


# Exotic and special purpose languages
registry.register(
    'pytuga',
    'ejudge.langs.pytuga.PytugaBuildManager',
    'ejudge.langs.pytuga.PytugaExecutionManager',
    extensions=['.pytg'],
    aliases=['pytg', 'pytugues', 'pytuguês']
)



from ejudge.build_manager import InterpretedLanguageBuildManager
from ejudge.execution_manager import InterpretedLanguageExecutionManager


class RubyBuildManager(InterpretedLanguageBuildManager):
    """
    Ruby builder.
    """

    source_extension = '.rb'

    def syntax_check(self):
        """
        Checks code for syntax errors.
        """

        #TODO: implement ruby syntax check.


class RubyExecutionManager(InterpretedLanguageExecutionManager):
    """
    Executes Ruby code.
    """

    interpreter_command = 'ruby'



import argparse

import ejudge
import iospec
from ejudge import __version__


def make_parser():
    """
    Creates parser object.
    """

    parser = argparse.ArgumentParser(
        description='Automatically grade the input file',
    )
    parser.add_argument('--version', '-v', action='version',
                        version='%(prog)s' + __version__)
    subparsers = parser.add_subparsers(
        title='subcommands',
        description='valid subcommands',
    )

    # ejudge run <source> <inputs>
    run_parser = subparsers.add_parser(
        'run',
        help='run program with the selected inputs'
    )
    run_parser.add_argument('file', help='input source code')
    run_parser.add_argument('inputs',
                            help='program inputs or IoSpec interaction')
    run_parser.set_defaults(func=command_run)

    # ejudge grade <source> <inputs>
    grade_parser = subparsers.add_parser(
        'grade',
        help='grade program with the selected inputs'
    )
    grade_parser.add_argument('file', help='input source code')
    grade_parser.add_argument('inputs', help='IoSpec interaction')
    grade_parser.set_defaults(func=command_grade)

    return parser


def command_run(args):
    """
    Implements "ejudge run <source> <inputs>" command.
    """

    source, lang = get_source_and_lang(args.file)
    if args.inputs.endswith('.iospec'):
        input_data = iospec.parse(args.inputs)
    else:
        with open(args.inputs) as F:
            input_data = F.read().splitlines()
    result = ejudge.run(source, input_data, lang=lang)
    print(result.source())


def command_grade(args):
    """
    Implements "ejudge grade <source> <inputs>" command.
    """

    source, lang = get_source_and_lang(args.file)
    input_data = iospec.parse(args.inputs)
    feedback = ejudge.grade(source, input_data, lang=lang)
    print(feedback.as_text())


def get_source_and_lang(path):
    """
    Return a tuple with (source, lang) for the given input file path.
    """

    with open(path) as F:
        source = F.read()
    lang = ejudge.registry.language_from_filename(path)
    return source, lang


def main():
    """
    Executes the main script.
    """

    parser = make_parser()
    args = parser.parse_args()
    try:
        func = args.func
    except AttributeError:
        print('Type `ejudge --help` for usage.')
    else:
        func(args)



from invoke import run, task

# Import some boilerplate tasks (see http://... for references)
from python_boilerplate.tasks import bump_version


# Add your tasks in here
# This task can be executed with ``invoke build --docs``
@task
def build(no_docs=False):
    """Build python package and docs"""

    run("python setup.py build")
    if not no_docs:
        run("python setup.py build_sphinx")


import decimal
import io
import traceback

from boxed.jsonbox import run as run_sandbox
from ejudge import registry
from ejudge.exceptions import BuildError
from ejudge.util import keep_cwd
from iospec import parse_string, TestCase, ErrorTestCase, IoSpec
from iospec.feedback import feedback as get_feedback, Feedback


def run(source, inputs, lang=None, *,
        timeout=None, raises=False, path=None, sandbox=True):
    """
    Run program with the given list of inputs and returns the corresponding
    :class:`iospec.IoSpec` instance with the results.

    Args:
        source (str or file)
            The source code for the test program
        inputs (sequence)
            A sequence of input strings. If input is a sequence of sequences,
            this function will perform multiple test cases. It can also be a
            IoSpec or a TestCase instance which are used to extract the
            necessary input strings.
        lang (str)
            The name for the source code language. See
            :func:`ejudge.graders.io.grade` for more details.
        timeout (float)
            A time limit for the entire run (in seconds). If this attribute is
            not given, the program will run without any timeout. This can be
            potentially dangerous if the input program has an infinite loop.
        sandbox (bool)
            Controls if code is run in sandboxed mode or not. Sandbox protection
            is the default behavior on supported platforms.
        raises (bool)
            If True, raise a BuildError if the build process fails. The default
            behavior is to return a IoSpec instance with a single ErrorTestCase
            with a type string of 'error-build'.
        path (str)
            The absolute file path for the input string or file object.

    Returns:
        A :class:`iospec.IoSpec` structure. If ``inputs`` is a sequence of
        strings, the resulting tree will have a single test case.
    """

    build_manager = registry.build_manager_from_path(lang, source, path,
                                                     is_sandboxed=sandbox)

    # Normalize inputs
    if isinstance(inputs, (IoSpec, TestCase)):
        inputs = inputs.inputs()
    else:
        if isinstance(inputs[0], str):
            inputs = [list(inputs)]
        else:
            inputs = [list(x) for x in inputs]

    # Execute
    with keep_cwd():
        if sandbox:
            imports = build_manager.get_modules()
            result = run_sandbox(
                run_from_lang,
                args=(build_manager.language, build_manager.source, inputs),
                kwargs={'raises': raises, 'timeout': timeout},
                imports=imports,
            )
            result = IoSpec.from_json(result)
        else:
            result = run_from_manager(
                build_manager,
                inputs,
                raises=raises,
                timeout=timeout
            )
    return result


def grade(source, iospec, lang=None, *,
          fast=True, path=None, raises=False, sandbox=False, timeout=None):
    """
    Grade the string of source code by comparing the results of all inputs and
    outputs in the given template structure.


    Args:
        source (str or file object)
            The source string for the code or a file object
        iospec (IOSpec parse tree)
            The expected template for correct answers.
        lang (str)
            Programming language for the given source code. Users can implement
            plugins to support additional languages or to override the default
            behavior or accepted languages.

            +-----------+------------------------------------------------------+
            | Value     | Description                                          |
            +===========+======================================================+
            | python    | For Python 3.x code. Default runner.                 |
            +-----------+------------------------------------------------------+
            | python2   | Executes in a separate Python 2 interpreter.         |
            +-----------+------------------------------------------------------+
            | tcc       | Compile C code with the tiny C compiler              |
            +-----------+------------------------------------------------------+
            | clang     | Compile C code with clang                            |
            +-----------+------------------------------------------------------+
            | gcc, c    | Compile C code with gcc                              |
            +-----------+------------------------------------------------------+

        sandbox (bool)
            If True, code will run in a sandboxed environment as the `nobody`
            user. It is necessary to have your system properly configured in
            order to do this.
        timeout (float)
            Maximum time (in seconds) for the complete test to run.

    Returns:
        A :class:`ejudge.Feedback` instance.
    """

    build_manager = registry.build_manager_from_path(lang, source, path,
                                                     is_sandboxed=sandbox)

    # Normalize inputs
    if isinstance(iospec, str):
        iospec = parse_string(iospec)
    if not iospec:
        raise ValueError('cannot grade an iospec that has no cases')

    with keep_cwd():
        kwargs = {'raises': raises, 'timeout': timeout, 'fast': fast}

        if sandbox:
            imports = build_manager.get_modules()
            language = build_manager.language
            manager_source = build_manager.source
            result = run_sandbox(
                grade_from_lang,
                args=(language, manager_source, iospec.to_json()),
                kwargs=kwargs,
                imports=imports,
            )
            result = Feedback.from_json(result)
        else:
            result = grade_from_manager(build_manager, iospec, **kwargs)
    return result


def get_answer_key_error(source, iospec, lang, **kwargs):
    """
    Compares the results of running the given source code with the given
    iospec and return the first test case which fails.

    Return None if no test case fails and the answer key is acceptable.
    """


def grade_from_manager(build_manager, iospec, raises, timeout, fast):
    """
    Grade manager instance by comparing it to the given iospec.
    """

    try:
        build_manager.build()
        error_message = None
    except BuildError as ex:
        if raises:
            raise
        error_message = str(ex)

    value = decimal.Decimal(1)
    feedback = None
    language = build_manager.language

    for answer_key in iospec:
        # We run each test case and compare results. If there was a build error
        # and raises is False, we create a new ErrorTestCase for each time
        # the program is supposed to run.
        inputs = answer_key.inputs()
        if error_message:
            case = ErrorTestCase.build(error_message=error_message)
        else:
            ctrl = registry.execution_manager(language, build_manager,
                                              inputs)
            try:
                if timeout:
                    ctrl.run_with_timeout(timeout)
                else:
                    case = ctrl.run()
            except Exception as ex:
                case = error_test_case(ex, ex.__traceback__)

            if not isinstance(case, TestCase):
                raise RuntimeError(
                    '%s .run() method did not return a TestCase: got a %s '
                    'instance' % (type(ctrl).__name__,  type(case).__name__),
                )

        # Compute feedback and compare with worst results
        curr_feedback = get_feedback(case, answer_key)

        if feedback is None:
            feedback = curr_feedback

        if curr_feedback.grade < value:
            feedback = curr_feedback
            value = curr_feedback.grade

            if value == 0 and fast:
                break

    return feedback


def grade_from_lang(lang, source, iospec, **kwargs):
    """
    A version of grade_from_manager() in which both the inputs and outputs
    can be converted to JSON.
    """

    iospec = IoSpec.from_json(iospec)
    build_manager = registry.build_manager(lang, source)
    result = grade_from_manager(build_manager, iospec, **kwargs)
    return result.to_json()


def error_test_case(exc, tb, limit=None):
    """
    Return an IoSpec data with a single ErrorTestCase.

    Construct the testcase from the given traceback.
    """

    out = io.StringIO()
    traceback.print_tb(tb, file=out)
    _, sep, tail = out.getvalue().partition('  File "main.py", line')
    message = ('Traceback (most recent call last)\n%s%s: %s' %
               (sep + tail, exc.__class__.__name__, exc))
    return ErrorTestCase.runtime(error_message=message)


def run_from_manager(build_manager, input_sets, *, raises, timeout):
    """
    Run command from given ExecutionManager.
    """

    try:
        build_manager.build()
    except BuildError as ex:
        if raises:
            raise
        return IoSpec([ErrorTestCase.build(error_message=str(ex))])

    data = []
    for inputs in input_sets:
        language = build_manager.language
        ctrl = registry.execution_manager(language, build_manager, inputs)
        try:
            if timeout:
                result = ctrl.run()
            else:
                result = ctrl.run_with_timeout(timeout)
            data.append(result)
        except Exception as ex:
            if raises:
                raise
            data.append(error_test_case(ex, ex.__traceback__))
    result = IoSpec(data)
    result.set_meta('lang', build_manager.language)
    return result


def run_from_lang(lang, src, inputs, *, raises, timeout):
    """
    Calls run_from_manager, but uses only JSON-encodable arguments.

    This function should be used in sandboxed environments.
    """

    build_manager = registry.build_manager(lang, src, is_sandboxed=True)
    result = run_from_manager(
        build_manager, inputs, raises=raises, timeout=timeout
    )
    return result.to_json()



import functools
import io
import multiprocessing
import os
import sys

from boxed.pinteract import Pinteract
from ejudge import builtins_ctrl
from ejudge.exceptions import MissingInputError
from ejudge.util import remove_trailing_newline_from_testcase, \
    timeout as run_with_timeout, format_traceback
from iospec import Out, In, types, SimpleTestCase, ErrorTestCase


class ExecutionManager:
    """
    Controls execution of a single test case.

    Args:
        build_manager:
            The BuildManager instance associated with this program.
        inputs:
            A list of lists of input strings.
    """

    source = property(lambda self: self.build_manager.source)
    is_sandboxed = property(lambda self: self.build_manager.is_sandboxed)

    def __init__(self, build_manager, inputs=()):
        self.build_manager = build_manager
        self.inputs = tuple(inputs)
        self.is_started = False
        self.is_closed = False

    def add_input(self, input_str):
        """
        Add an input string to the end of registered inputs.
        """

        self.inputs += (input_str,)

    def add_inputs(self, seq):
        """
        Add a sequence of input strings to the end of registered inputs.
        """

        self.inputs += tuple(seq)

    def run(self):
        """
        Run program and return a list of In/Out elements interactions.
        """

        if self.is_closed or self.build_manager.is_closed:
            raise RuntimeError(
                'Program already run. Cannot run it again.'
            )
        if self.is_started:
            raise RuntimeError(
                'Program already started execution. Please create another '
                'ExecutionManager instance.'
            )
        self.start()
        result = self.interact()
        self.end()
        return remove_trailing_newline_from_testcase(result)

    def run_with_timeout(self, timeout):
        """
        Run execution with a given timeout.
        """

        return run_with_timeout(self.run, timeout=timeout)

    def interact(self):
        """
        Interact with the program by using all input strings.

        Return a IoSpec.TestCase instance with the results of interaction.
        """

        raise NotImplementedError

    def start(self):
        """
        Executed to start program execution.
        """

        if not self.build_manager.is_built:
            self.build_manager.build()
        self.is_started = True

    def end(self):
        """
        Executed after execution ends. May be necessary to clean state.
        """

        self.is_closed = True


class IntegratedExecutionManager(ExecutionManager):
    """
    Execution manager for Python or other languages integrated into the Python
    interpreter.
    """
    __print = staticmethod(print)
    __input = staticmethod(input)

    def __init__(self, build_manager, inputs):
        super().__init__(build_manager, inputs)
        self.interaction = []
        self.is_isolated = False

    def interact_integrated(self):
        builtins_ctrl.update(self.builtins())
        try:
            if self.build_manager.locals is None:
                locals_dic = None
            else:
                locals_dic = dict(self.build_manager.locals)
            globals_dic = dict(self.build_manager.globals or {})
            try:
                self.exec(globals_dic, locals_dic)
                result = SimpleTestCase(self.interaction)
            except Exception as ex:
                error = format_traceback(ex, self.source)
                result = ErrorTestCase.runtime(
                    self.interaction,
                    error_message=error,
                )
        finally:
            builtins_ctrl.restore()

        return result

    def interact(self):
        if self.is_sandboxed:
            return self.interact_integrated()
        else:
            # We spawn and immediately join the child process in order to
            # execute the integrated manager in a separate environment. This
            # isolation is important in execution environments that introduce
            # global state to the interpreter.
            #
            # Since python do not really prevent scripts from introducing global
            # state, we always isolate execution this way to prevent any
            # potentially dangerous global state to leak into the main
            # interpreter process.
            queue = multiprocessing.Queue()
            process = multiprocessing.Process(
                target=integrated_manager_interact,
                kwargs={
                    'queue': queue,
                    'inputs': self.inputs,
                    'build_manager': self.build_manager,
                    'execution_manager_class': type(self),
                }
            )
            process.start()
            process.join()
            return queue.get()

    # noinspection PyShadowingBuiltins
    def exec(self, globals, locals):
        """
        Execute code with the given locals and globals.
        """

        code = compile(self.source, 'main.py', 'exec')

        if locals is None:
            exec(code, globals)
        else:
            exec(code, globals, locals)

    def builtins(self):
        """
        Return a dictionary with builtin functions replacements.

        The default implementation just replaces the builtin print() and input()
        functions.
        """
        consumed_inputs = list(self.inputs)[::-1]

        @functools.wraps(self.__print)
        def print(*args, sep=' ', end='\n', file=None, flush=False):
            if not (file is None or file is sys.stdout):
                self.__print(*args, sep=sep, end=end, file=file, flush=flush)
            else:
                file = io.StringIO()
                self.__print(*args, sep=sep, end=end, file=file, flush=flush)
                data = file.getvalue()

                if self.interaction and isinstance(self.interaction[-1], Out):
                    self.interaction[-1] += Out(data)
                else:
                    self.interaction.append(Out(data))

        @functools.wraps(self.__input)
        def input(prompt=None):
            if prompt is not None:
                print(prompt, end='')
            if consumed_inputs:
                result = consumed_inputs.pop()
                self.interaction.append(In(result))
                return result
            else:
                raise MissingInputError('not enough inputs')

        return {'print': print, 'input': input}


class PInteractExecutionManager(ExecutionManager):
    """
    Execution manager for languages executed outside the main Python 3
    interpreter.
    """

    shell_args = None

    def interact(self):
        return self.run_pinteract(self.get_shell_args())

    def run_pinteract(self, shell_args):
        """
        Run script as a subprocess and gather results of execution.

        This is a generic implementation. Specific languages or runtimes are
        supported by fixing the shellargs argument of this function.
        This function uses the PInteract() object for communication with
        scripts.

        This function may change the current working path.
        """

        def append_non_empty_output():
            data = process.receive()
            if data:
                result.append(types.Out(data))

        # Execute script in the tempdir and than go back once execution has
        # finished
        result = SimpleTestCase()
        current_dir = os.getcwd()

        try:
            os.chdir(self.build_manager.build_path)
            process = Pinteract(shell_args)

            # Fetch all In/Out strings
            append_non_empty_output()
            for idx, inpt in enumerate(self.inputs):
                try:
                    process.send(inpt)
                    result.append(types.In(inpt))
                except RuntimeError as ex:
                    # Early termination: we still have to decide if an specific
                    # early termination error should exist.
                    #
                    # The default behavior is just to send a truncated
                    # IoTestCase
                    if process.is_dead():
                        missing = self.inputs[idx:]
                        missing_str = '\n'.join('    ' + x for x in missing)
                        msg = ('process closed with consuming all inputs. '
                               'List of unused inputs:\n')

                        return ErrorTestCase.earlytermination(
                            list(result),
                            error_message=msg + '\n' + missing_str
                        )

                    # This clause is just a safeguard. We don't expect to ever
                    # get here
                    return ErrorTestCase.runtime(
                        list(result),
                        error_message='An internal error occurred while trying '
                                      'to interact with the script: %s.' % ex
                    )
                append_non_empty_output()
        finally:
            if not self.is_sandboxed:
                os.chdir(current_dir)

        # Finish process
        error_ = process.finish()
        assert not any(error_), error_
        return result

    def get_shell_args(self):
        """
        Return the arguments passed to the executable object.
        """

        if self.shell_args is None:
            raise NotImplementedError(
                'Subclasses must either define the .shell_args attribute or '
                'override the .get_shell_args() method.'
            )
        if isinstance(self.shell_args, str):
            import shlex
            return shlex.split(self.shell_args)
        return list(self.shell_args)


class CompiledLanguageExecutionManager(PInteractExecutionManager):
    """
    Base execution manager for compiled languages.

    It assumes that the compiled executable is named "main.exe" and is inside
    the build directory.
    """

    shell_args = ['./main.exe']


class InterpretedLanguageExecutionManager(PInteractExecutionManager):
    """
    Base execution manager for interpreted languages.

    It assumes the main interpreter will just execute the source file.
    """

    interpreter_command = None

    def get_interpreter_command(self):
        """
        Return the command to execute the interpreter.
        """

        if self.interpreter_command is None:
            raise NotImplementedError(
                'Subclasses must either supply the .interpreter_command '
                'string attribute or override the get_interpreter_command() '
                'method.'
            )

        source_name = self.build_manager.get_source_filename()
        return [self.interpreter_command, source_name]

    def get_shell_args(self):
        if self.shell_args is None:
            return self.get_interpreter_command()
        else:
            return self.shell_args


# Interact with integrated manager as the target function in a subprocess
# execution.
def integrated_manager_interact(queue, inputs, build_manager,
                                execution_manager_class):
    exc_manager = execution_manager_class(build_manager, inputs)
    result = exc_manager.interact_integrated()
    queue.put(result)



"""
Import builtin functions for easy recovery when a script or a preparation
routine messes with Python's __builtins__.
"""
import builtins

# Save this for later
__original__ = vars(builtins)
_print = print
_input = input


def update(dic=None, **kwds):
    """
    Update the builtins module with the given values
    """

    dic = dict(dic or {})
    dic.update(kwds)
    for k, v in dic.items():
        setattr(builtins, k, v)


def restore():
    """
    Restores builtins to original state
    """

    for k, v in vars(builtins).items():
        if k not in __original__:
            delattr(builtins, k)
    for k, v in __original__.items():
        setattr(builtins, k, v)
    builtins.print = _print
    builtins.input = _input



import contextlib
import io
import os
import sys
import signal
import traceback
from threading import Thread


#
# Special IO functions for python script interactions
#
from iospec import types

_print_func = print
_stdout = sys.stdout


def real_print(*args, **kwargs):
    """Print function for the C's stdout."""

    kwargs.setdefault('file', _stdout)
    _print_func(*args, **kwargs)


def capture_print(*args, **kwds):
    """A print-like function that return the formatted string instead of
    printing it on screen"""

    out, err = sys.stdout, sys.stderr
    out_io = sys.stdout = sys.stderr = io.StringIO()
    try:
        _print_func(*args, **kwds)
    finally:
        sys.stdout, sys.stderr = out, err
    return out_io.getvalue()


#
# Execution with time limits
#
def __timeout_handler(signum, frame):
    """Helper function for the timeout() function."""

    raise TimeoutError()


def timeout(func, args=(), kwargs={}, timeout=1.0, threading=True, raises=True):
    """
    Execute callable `func` with timeout. If timeout is None or zero,
    ignores any timeout exceptions.

    If timeout exceeds, raises a TimeoutError.
    """

    if not timeout or not 1 / timeout:
        return func(*args, **kwargs)

    if threading:
        result = []
        exceptions = []

        def target():
            try:
                result.append(func(*args, **kwargs))
            except Exception as ex:
                exceptions.append(ex)

        thread = Thread(target=target)
        thread.start()
        thread.join(timeout)
        if thread.is_alive():
            if raises:
                raise TimeoutError
            else:
                return None
        else:
            try:
                return result.pop()
            except IndexError:
                raise exceptions.pop()
    else:
        signal.signal(signal.SIGALRM, __timeout_handler)
        signal.alarm(timeout)
        try:
            result = func(*args, **kwargs)
        except TimeoutError:
            if raises:
                raise
            else:
                return None
        finally:
            signal.alarm(0)

        return result


#
# Lazy evaluation
#
class lazy(object):
    """Lazy accessor"""

    def __init__(self, func):
        self.func = func

    def __get__(self, obj, cls=None):
        if obj is None:
            return self
        value = self.func(obj)
        setattr(obj, self.func.__name__, value)
        return value


class pushbackiter:
    """A simple push back iterator: push()'d items are restored to iteration.

    Example
    -------

    >>> it = pushbackiter([1, 2, 3]);
    >>> it.push(0); list(it)
    [0, 1, 2, 3]
    """

    def __init__(self, it):
        self.iter = iter(it)
        self.buffer = []

    def __iter__(self):
        return self

    def __next__(self):
        buffer = self.buffer
        while buffer:
            return buffer.pop()
        return next(self.iter)

    def push(self, value):
        """Send a value to the beginning of iteration."""

        self.buffer.append(value)


@contextlib.contextmanager
def keep_cwd():
    """
    Context manager that restores the cwd at the beginning of execution when it
    exits the with block."""

    currpath = os.getcwd()
    yield
    os.chdir(currpath)


def remove_trailing_newline_from_testcase(case):
    """
    Remove a trailing newline from the last output node.
    """

    if case is None:
        return None

    if case:
        last = case[-1]
        if isinstance(last, types.Out) and last.endswith('\n'):
            last.data = str(last[:-1])
    return case


def format_traceback(ex, source):
    """
    Creates a error message string from an exception with a __traceback__
    value. Usually this requires that this function must be executed inside
    the except block that caught the exception.
    """

    ex_name = type(ex).__name__
    messages = []
    code_lines = source.splitlines()
    tb = ex.__traceback__

    tb_list = reversed(traceback.extract_tb(tb))
    for (filename, lineno, func_name, text) in tb_list:
        if 'ejudge' in filename:
            break
        if filename == '<string>':
            text = code_lines[lineno - 1].strip()
        messages.append((filename, lineno, func_name, text))

    messages.reverse()
    messages = traceback.format_list(messages)
    messages.insert(0, 'Traceback (most recent call last)')
    messages.append('%s: %s' % (ex_name, ex))
    return '\n'.join(messages)



# -*- coding: utf8 -*-
import os

from setuptools import setup, find_packages

# Meta information
name = 'ejudge'
author = 'Fábio Macêdo Mendes'
version = open('VERSION').read().strip()
dirname = os.path.dirname(__file__)

# Save version and author to __meta__.py
with open(os.path.join(dirname, 'src', name, '__meta__.py'), 'w') as F:
    F.write('__version__ = %r\n__author__ = %r\n' % (version, author))

setup(
    # Basic info
    name=name,
    version=version,
    author=author,
    author_email='fabiomacedomendes@gmail.com',
    url='http://github.com/fabiommendes/ejudge',
    description='A simple ejudge for Python. Supports several languages and '
                'sandboxing.',
    long_description=open('README.rst').read(),

    # Classifiers (see https://pypi.python.org/pypi?%3Aaction=list_classifiers)
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License (GPL)',
        'Operating System :: POSIX',
        'Programming Language :: Python',
        'Topic :: Education',
        'Topic :: Software Development :: Libraries',
    ],

    # Packages and depencies
    package_dir={'': 'src'},
    packages=find_packages('src'),
    install_requires=[
        'psutil',
        'pexpect',
        'iospec>=0.3',
        'boxed>=0.3'
    ],
    extras_require={
        'testing': ['pytest'],
    },

    # Scripts
    entry_points={
       'console_scripts': ['ejudge = ejudge.__main__:main'],
    },

    # Other configurations
    zip_safe=False,
    platforms='any',
    test_suite='%s.test.test_%s' % (name, name),
)



import importlib
import os

from ejudge.build_manager import BuildManager


class LanguageRegistry:
    """
    Registry that maps languages to managers.
    """

    def __init__(self):
        self._extensions = {}
        self._execution_managers = {}
        self._build_managers = {}

    def _register_extension(self, ext, language, force=False):
        """
        Register extension to the given language.
        """

        if ext.startswith('.'):
            ext = ext[1:]

        extensions = self._extensions
        if force and ext in extensions and ext != extensions[language]:
            raise RuntimeError(
                'extension %r is already registered to .%r' % (ext, language)
            )
        extensions[ext] = language

    def register(self, language, build, execution,
                 force=False,
                 extensions=(),
                 aliases=()):
        """
        Register execution/build manager classes for language.
        """

        for ext in extensions:
            self._register_extension(ext, language, force=force)

        self._execution_managers[language] = execution
        self._build_managers[language] = build
        for alias in aliases:
            self._execution_managers[alias] = execution
            self._build_managers[alias] = build

    def build_manager(self, language, source, **kwargs):
        """
        Return a build manager instance for the given language string.

        It passes the source code instance and any other additional arguments
        passed to this function to the class constructor.
        """

        manager = self._build_managers[language]

        if isinstance(manager, str):
            mod_name, _, manager_name = manager.rpartition('.')
            mod = importlib.import_module(mod_name)
            manager = getattr(mod, manager_name)

        manager = manager(source, **kwargs)
        manager.language = language
        return manager

    def execution_manager(self, language, build_manager, inputs=(),
                          **kwargs):
        """
        Return an execution manager instance for the given language string.

        It passes the build_manager instance and any other additional arguments
        passed to this function to the class constructor.
        """

        assert isinstance(build_manager, BuildManager), build_manager
        manager_class = self.execution_manager_class(language)
        manager = manager_class(build_manager, inputs, **kwargs)
        manager.language = language
        return manager

    def execution_manager_class(self, language):
        """
        Return the execution manager class associated with the given language.
        """

        manager_class = self._execution_managers[language]

        if isinstance(manager_class, str):
            mod_name, _, manager_name = manager_class.rpartition('.')
            mod = importlib.import_module(mod_name)
            manager_class = getattr(mod, manager_name)

        return manager_class

    def language_from_extension(self, ext):
        """
        Return a language string from the given extension.
        """

        if ext.startswith('.'):
            ext = ext[1:]
        try:
            return self._extensions[ext]
        except KeyError:
            raise ValueError('unknown extension: %r' % ext)

    def language_from_filename(self, path):
        """
        Return the language string from a given filename.
        """

        ext = os.path.splitext(path)[1] or '.'
        return self.language_from_extension(ext)

    def language_from_source(self, file, path=None):
        """
        Return the language from either an explicit path or from a file object
        .name attribute.
        """

        return self.language_from_filename(getattr(file, 'name', path))

    def build_manager_from_path(self, lang, source, path, **kwargs):
        """
        Uses path information to load the most appropriate BuildManager.

        It accepts either a source string or a source file object.
        """

        if lang is None:
            lang = self.language_from_source(source, path)

        if not isinstance(source, str):
            source = source.read()

        return self.build_manager(lang, source, **kwargs)


registry = LanguageRegistry()



try:
    # Reuse exception from IoSpec.
    from iospec.exceptions import BuildError
except ImportError:
    class BuildError(Exception):
        """
        Exception raised when the build process is unsuccessful.
        """


class MissingInputError(RuntimeError):
    """
    Error raised when asked for a non-existing input
    """


class EarlyTerminationError(RuntimeError):
    """
    Error raised when program finishes without consuming all inputs.
    """
